@@ -91,6 +91,12 @@ type UserGrafanaConfig struct {
 	StaticHeaders map[string]string `json:"static_headers"`
 }
 
+type UserGrafanaConfigStatus struct {
+	Hash      string `json:"configuration_hash"`
+	CreatedAt int64  `json:"created"`
+	Promoted  bool   `json:"promoted"`
+}
+
 func (gc *UserGrafanaConfig) Validate() error {
 	if gc.Hash == "" {
 		return errors.New("no hash specified")
@@ -447,7 +453,7 @@ func (am *MultitenantAlertmanager) SetUserGrafanaConfig(w http.ResponseWriter, r
 	}
 
 	cfgDesc := alertspb.ToGrafanaProto(cfg.GrafanaAlertmanagerConfig.original, userID, cfg.Hash, cfg.CreatedAt, cfg.Default, cfg.Promoted, cfg.ExternalURL, cfg.SmtpFrom, cfg.StaticHeaders, smtpConfig)
-	if err := validateUserGrafanaConfig(logger, cfgDesc, am.limits, userID); err != nil {
+	if err := am.validateUserGrafanaConfig(logger, cfgDesc, am.limits, userID); err != nil {
 		level.Error(logger).Log("msg", errValidatingConfig, "err", err.Error())
 		w.WriteHeader(http.StatusBadRequest)
 		util.WriteJSONResponse(w, errorResult{Status: statusError, Error: fmt.Sprintf("%s: %s", errValidatingConfig, err.Error())})
@@ -487,8 +493,41 @@ func (am *MultitenantAlertmanager) DeleteUserGrafanaConfig(w http.ResponseWriter
 	util.WriteJSONResponse(w, successResult{Status: statusSuccess})
 }
 
+func (am *MultitenantAlertmanager) GetGrafanaConfigStatus(w http.ResponseWriter, r *http.Request) {
+	logger := util_log.WithContext(r.Context(), am.logger)
+
+	userID, err := tenant.TenantID(r.Context())
+	if err != nil {
+		level.Error(logger).Log("msg", errNoOrgID, "err", err.Error())
+		w.WriteHeader(http.StatusUnauthorized)
+		util.WriteJSONResponse(w, errorResult{Status: statusError, Error: fmt.Sprintf("%s: %s", errNoOrgID, err.Error())})
+		return
+	}
+
+	cfg, err := am.store.GetGrafanaAlertConfig(r.Context(), userID)
+	if err != nil {
+		if errors.Is(err, alertspb.ErrNotFound) {
+			w.WriteHeader(http.StatusNotFound)
+			util.WriteJSONResponse(w, errorResult{Status: statusError, Error: err.Error()})
+		} else {
+			w.WriteHeader(http.StatusInternalServerError)
+			util.WriteJSONResponse(w, errorResult{Status: statusError, Error: err.Error()})
+		}
+		return
+	}
+
+	util.WriteJSONResponse(w, successResult{
+		Status: statusSuccess,
+		Data: &UserGrafanaConfigStatus{
+			Hash:      cfg.Hash,
+			CreatedAt: cfg.CreatedAtTimestamp,
+			Promoted:  cfg.Promoted,
+		},
+	})
+}
+
 // ValidateUserGrafanaConfig validates the Grafana Alertmanager configuration.
-func validateUserGrafanaConfig(logger log.Logger, cfg alertspb.GrafanaAlertConfigDesc, limits Limits, user string) error {
+func (am *MultitenantAlertmanager) validateUserGrafanaConfig(logger log.Logger, cfg alertspb.GrafanaAlertConfigDesc, limits Limits, user string) error {
 	// We don't have a valid use case for empty configurations. If a tenant does not have a
 	// configuration set and issue a request to the Alertmanager, we'll a) upload an empty
 	// config and b) start an Alertmanager instance for them if a fallback
@@ -498,7 +537,7 @@ func validateUserGrafanaConfig(logger log.Logger, cfg alertspb.GrafanaAlertConfi
 	}
 
 	// Perform a similar flow of transformations that would happen in the Alertmanager on Sync & Apply.
-	grafanaConfig, err := createUsableGrafanaConfig(logger, cfg, "")
+	grafanaConfig, err := am.amConfigFromGrafanaConfig(cfg)
 	if err != nil {
 		return err
 	}
